import shapefile
import math
from os import path
import numpy as np
import sys
import bisect

class SHPC_class:
    def __init__(self, SHPC_dir, def_orient):
        """def_orient, for default orientation, is used to define d_init and
        n_dates. Since we are going to open files, we might as
        well ask the user which orientation will be useful.

        """

        self.dir = SHPC_dir
        self.def_orient = def_orient
        fname = path.join(SHPC_dir, "n_slices.txt")

        try:
            with open(fname) as f: self.n_slices = int(f.read())
        except FileNotFoundError:
            print(fname, "not found. Creating it...")
            self.n_slices = 0

            while True:
                slice_dir = path.join(SHPC_dir, f"Slice_{self.n_slices}")
                if not path.isdir(slice_dir): break
                self.n_slices += 1

            if self.n_slices == 0:
                print("util_eddies:SHPC_class:init:No slice found")
                sys.exit(1)
            else:
                with open(fname, "w", encoding="utf-8") as f:
                    f.write(f"{self.n_slices}\n")

        self._readers = [{"Anticyclones": {}, "Cyclones": {}}
                         for i in range(self.n_slices)]
        self.d_init = []

        for i_slice, reader in enumerate(self._readers):
            my_shapefile = path.join(SHPC_dir, f"Slice_{i_slice}", def_orient,
                                     "extremum")
            try:
                reader[def_orient]["extremum"] = shapefile.Reader(my_shapefile)
            except shapefile.ShapefileException:
                # Use the other orientation for this slice:
                orientation = other_orient(def_orient)
                my_shapefile = path.join(SHPC_dir, f"Slice_{i_slice}",
                                         orientation, "extremum")
                reader[orientation]["extremum"] = shapefile.Reader(my_shapefile)
            else:
                orientation = def_orient

            d_init_slice = reader[orientation]["extremum"].record(0).date
            self.d_init.append(d_init_slice)

        self._n_dates = np.zeros(self.n_slices, dtype = int)
        self._ishape_last = [{"Anticyclones": None, "Cyclones": None}
                             for i in range(self.n_slices)]

        # Number of eddies at each date:
        self._e_max = [{"Anticyclones": None, "Cyclones": None}
                             for i in range(self.n_slices)]

        self._n_shapes = [{"Anticyclones": 0, "Cyclones": 0}
                         for i in range(self.n_slices)]

    def get_n_dates(self, i_slice):
        """Return the number of dates in a given slice, including missing
        dates. That is last date in the slice minus first date in the
        slice plus one.

        """

        if self._n_dates[i_slice] == 0:
            # Find out the number of dates in the slice:
            if self._ishape_last[i_slice]["Anticyclones"] is not None:
                self._n_dates[i_slice] \
                    = self._ishape_last[i_slice]["Anticyclones"].size
            elif self._ishape_last[i_slice]["Cyclones"] is not None:
                self._n_dates[i_slice] \
                    = self._ishape_last[i_slice]["Cyclones"].size
            else:
                try:
                    self._n_dates[i_slice] \
                        = self.get_ishape_last(i_slice, self.def_orient).size
                except AttributeError:
                    # Try the other orientation:
                    orientation = other_orient(self.def_orient)
                    self._n_dates[i_slice] \
                        = self.get_ishape_last(i_slice, orientation).size

        return self._n_dates[i_slice]

    def get_ishape_last(self, i_slice, orientation):
        """Can return None if ishape_last.txt does not exist and cannot be
        created.

        """

        if self._ishape_last[i_slice][orientation] is None:
            fname = path.join(self.dir, f"Slice_{i_slice}", orientation,
                              "ishape_last.txt")
            print("Reading", fname, "...")

            try:
                self._ishape_last[i_slice][orientation] \
                    = np.loadtxt(fname, dtype = int, ndmin = 1)
            except OSError:
                print("Could not read", fname)
                print("Creating it...")
                self._create_ishape_last(i_slice, orientation)

            if self._ishape_last[i_slice][orientation] is not None:
                assert self._ishape_last[i_slice][orientation].size != 0
                self._e_max[i_slice][orientation] \
                    = np.ediff1d(self._ishape_last[i_slice][orientation],
                                 to_begin = self._ishape_last[i_slice]\
                                 [orientation][0] + 1)

        return self._ishape_last[i_slice][orientation]

    def get_reader(self, i_slice, layer, orientation):
        if layer not in self._readers[i_slice][orientation]:
            my_shapefile = path.join(self.dir, f"Slice_{i_slice}", orientation,
                                     layer)
            print("Opening", my_shapefile, "...")

            try:
                self._readers[i_slice][orientation][layer] \
                    = shapefile.Reader(my_shapefile)
            except shapefile.ShapefileException:
                print("Not found")
                self._readers[i_slice][orientation][layer] = None

        return self._readers[i_slice][orientation][layer]

    def get_n_shapes(self, i_slice, orientation):
        if self._n_shapes[i_slice][orientation] == 0:
            reader = self.get_reader(i_slice, layer = "extremum",
                                     orientation = orientation)
            self._n_shapes[i_slice][orientation] = len(reader)

        return self._n_shapes[i_slice][orientation]

    def _create_ishape_last(self, i_slice, orientation):
        """Can fail if it cannot get a reader."""

        last_date = self.d_init[i_slice]
        date_ishape = [] # list of tuples (date, ishape)
        reader = self.get_reader(i_slice, "extremum", orientation)

        if reader is None:
            print('Could not create "ishape_last.txt".')
        else:
            for ishape, rec in enumerate(reader.iterRecords(fields = ["date"])):
                if rec.date != last_date:
                    date_ishape.append((last_date, ishape - 1))
                    last_date = rec.date

            date_ishape.append((last_date, ishape))
            date_ishape = np.array(date_ishape)

            # Replace date by repetition of date factor:
            date_ishape[:- 1, 0] = np.ediff1d(date_ishape[:, 0])
            date_ishape[- 1, 0] = 1

            ishape_last = []

            for repet_factor, ishape in date_ishape:
                ishape_last.extend(repet_factor * [ishape])

            self._ishape_last[i_slice][orientation] = np.array(ishape_last)
            fname = path.join(self.dir, f"Slice_{i_slice}", orientation,
                              "ishape_last.txt")
            np.savetxt(fname, self._ishape_last[i_slice][orientation],
                       fmt = "%.0d")

    def get_slice(self, date_index):
        i_slice = bisect.bisect(self.d_init, date_index) - 1
        assert i_slice >= 0
        return i_slice

    def comp_ishape(self, date, eddy_index, i_slice, orientation):
        """Compute the location in the shapefiles. Returns None if ishape_last
        was not found and could not be created. Crashes if date is not
        in i_slice. Returns None if eddy_index is greater than the
        maximum value.

        """

        ishape_last = self.get_ishape_last(i_slice, orientation)

        if ishape_last is None:
            ishape = None
        else:
            k = date - self.d_init[i_slice]
            assert k >= 0
            assert 1 <= eddy_index

            if eddy_index <= self._e_max[i_slice][orientation][k]:
                if k == 0:
                    ishape = eddy_index - 1
                else:
                    # k >= 1
                    ishape = ishape_last[k - 1] + eddy_index
            else:
                ishape = None

        return ishape

    def ishape_range(self, date, i_slice, orientation):
        """Returns an empty list if ishape_last was not found and could not be
        created. Crashes if date is not in i_slice.

        """

        ishape_start = self.comp_ishape(date, 1, i_slice, orientation)

        if ishape_start is None:
            return []
        else:
            ishape_stop = self._ishape_last[i_slice][orientation]\
                [date - self.d_init[i_slice]] + 1
            return range(ishape_start, ishape_stop)

def in_window(point, window):
    llcrnrlon, llcrnrlat, urcrnrlon, urcrnrlat = window
    longit = point[0] + math.ceil((llcrnrlon - point[0]) / 360) * 360
    # (in [llcrnrlon, llcrnrlon + 360[)

    return longit <= urcrnrlon and llcrnrlat <= point[1] <= urcrnrlat

def other_orient(orientation):
    if orientation == "Cyclones":
        return "Anticyclones"
    else:
        return "Cyclones"
